/** ZCT/globals.c--Created 070105 **/
#include "zct.h"
#include "functions.h"

int singular_extension;
int singular_margin;
/* board representation */
BOARD board;
int process_count;

/* various board data */
int pawn_dir[2] = { 8, -8 };
CASTLE_RIGHTS castle_rights_mask[64] = /* which castling flags each square affects */
{
	~CASTLE_WQ, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_WB, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_WK,
	~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO,
	~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO,
	~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO,
	~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO,
	~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO,
	~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO,
	~CASTLE_BQ, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_BB, ~CASTLE_NO, ~CASTLE_NO, ~CASTLE_BK
};
BITBOARD mask_bb[64];
BITBOARD pawn_caps_bb[2][64];
BITBOARD knight_moves_bb[64];
BITBOARD king_moves_bb[64];
BITBOARD in_between_bb[64][64];
BITBOARD castle_ks_mask[2];
BITBOARD castle_qs_mask[2];
HASHKEY zobrist_piece[2][6][64];
HASHKEY zobrist_castle[16];
HASHKEY zobrist_ep[65];
HASHKEY zobrist_stm;

/* interface data */
PROTOCOL protocol;
NOTATION notation = SAN;
FILE *input_stream;
FILE *output_stream;
FILE *log_stream;
BOOL source = FALSE;
char input_buffer[BUFSIZ] = "";
BOOL use_book = TRUE;
BOOL post;
int output_limit = 10000;
BOOL hard;
int ping = 0;
BOOL ics_mode = FALSE;

/* statistics and search data */
ENGINE_STATE engine_state;
COLOR zct_side;
BOOL game_over;
RESULT game_result;
MOVE ponder_move;
int search_age;
int max_depth = 0;
int max_nodes = 0;
int current_depth;
GAME_ENTRY *root_entry;
ROOT_MOVE *next_root_move;
ROOT_MOVE root_move_list[MAX_ROOT_MOVES];
int root_move_count;
int root_pv_counter;
VALUE last_root_score;
VALUE best_score_by_depth[MAX_PLY];
int pv_changes_by_depth[MAX_PLY];
BITBOARD nodes;
BITBOARD q_nodes;
int max_depth_reached;
int nodes_per_check;
int nodes_until_check;
int sb_id = 0;
BITBOARD pv_nodes;
BITBOARD fail_high_nodes;
BITBOARD fail_low_nodes;
BITBOARD pv_first;
BITBOARD fail_high_first;
BITBOARD hash_probes;
BITBOARD hash_hits;
BITBOARD hash_cutoffs;
BITBOARD hash_entries_full;
BITBOARD pawn_hash_probes;
BITBOARD pawn_hash_hits;
BITBOARD eval_hash_probes;
BITBOARD eval_hash_hits;
BITBOARD qsearch_hash_probes;
BITBOARD qsearch_hash_hits;
unsigned int history_counter=1;
unsigned int history_table[2][4096];
MOVE killer_move[MAX_PLY][2];
MOVE counter_move[2][4096][2];
int history_threshold = 60;
int check_extension = 3;
int one_rep_extension = 4;
int threat_extension = 4;
int passed_pawn_extension = 3;
int check_extensions_done;
int one_rep_extensions_done;
int threat_extensions_done;
int passed_pawn_extensions_done;
int singular_extensions_done;
BITBOARD hash_size = 1 * HASH_MB;
unsigned int qsearch_hash_size = 512 * HASH_KB;
unsigned int pawn_hash_size = 512 * HASH_KB;
unsigned int eval_hash_size = 1 * EVAL_HASH_MB;
HASH_ENTRY *hash_table = NULL;
HASH_ENTRY *qsearch_hash_table = NULL;
PAWN_HASH_ENTRY *pawn_hash_table = NULL;
EVAL_HASH_ENTRY *eval_hash_table = NULL;

/* other constants */
const char color_str[2][6] = { "white", "black" };
const char piece_char[2][7] = { "PNBRQK", "pnbrqk" };
const char piece_str[6][7] = { "pawn", "knight", "bishop", "rook", "queen", "king" };
const char result_str[][64] =
{
	"*", "1-0 {White mates}", "0-1 {Black mates}", "0-1 {White resigns}", "1-0 {Black resigns}",
	"1/2-1/2 {Stalemate}", "1/2-1/2 {Fifty move rule}", "1/2-1/2 {Three repetitions}",
	"1/2-1/2 {Draw by insufficient material", "1/2-1/2 {Draw by mutual agreement}"
};
