/** ZCT/ponder.c--Created 102506 **/
#include "zct.h"
#include "functions.h"
#include "globals.h"

/**
ponder():
This controls the ponder search, which is used when the opponent is on move. A move is selected
that is likely for the opponent to play, and then ZCT thinks as if this move has been played. If the
opponent plays this move, then we have saved time that can be used later.
Created 102506; last modified 011308
**/
void ponder(void)
{
start:
	/* Right now we simply take the second move from the last principal variation as the ponder move.
		If there isn't one, we just skip pondering. */
	if (ponder_move == NULL_MOVE)
		return;
	if (protocol == XBOARD)
		print("Hint: %M\n", ponder_move);
	else
	{
		prompt();
		print("%M (pondering)\n", ponder_move);
	}

	make_move(ponder_move);
	/* Check the result to see if the ponder move ends the game, and don't ponder if it does. */
	if (check_result(FALSE))
	{
		unmake_move();
		return;
	}
	engine_state = PONDERING;
	generate_root_moves();
	search_root();
	/* Engine_state is set to NORMAL on a ponder hit. We must send the move we made and update the clocks. */
	if (engine_state == NORMAL)
	{
		engine_state = IDLE;
		if (protocol == XBOARD)
			print("move ");
		else
			prompt();
		print("%M\n", board.pv_stack[0][0]);
		make_move(board.pv_stack[0][0]);
		update_clocks();
		if (ping)
		{
			print("pong %i\n", ping);
			ping = 0;
		}
		goto start;
	}
	unmake_move();
	engine_state = IDLE;
}
