/** ZCT/searchroot.c--Created 091505 **/
#include "zct.h"
#include "functions.h"
#include "globals.h"
#include "smp.h"
#include "eval.h"

int idle_time;

/**
search_root():
Do an alpha-beta search at the root. This routine also implements iterative deepening.
Created 091505; last modified 020208
**/
void search_root(void)
{
#ifdef SMP
	int x;
#endif
	VALUE alpha;
	VALUE beta;
	VALUE value;
	BITBOARD last_nodes;
	ROOT_MOVE *move;

#ifdef SMP
	smp_block[0].idle = FALSE;
#endif
	init_stats();
	root_pv_counter = 0;
	root_entry = board.game_entry;
	if (engine_state != PONDERING)
		ponder_move = NULL_MOVE;
	
	/* Probe the book. If we get a hit, we don't have to search. */
	if (use_book && engine_state != ANALYZING && engine_state != PONDERING &&
		engine_state != INFINITE && book_probe())
		return;

	start_time();
	set_time_limit();
#ifdef SMP
	start_child_processors();
//   if (!egbb_is_loaded || pop_count(board.occupied_bb) > 5) /* TODO: swindle mode */
//		;
#endif

	if (post && protocol != UCI)
		print("Depth          Time     Score   PV (nodes)\n");
	/* The main iterative deepening loop. */
	for (current_depth = PLY; current_depth <= MAX_PLY * PLY && !stop_search(); current_depth += PLY)
	{
//smp_data->current_depth = current_depth;
		alpha = -MATE;
		beta = MATE;
		next_root_move = &root_move_list[0];
		while ((move = select_root_move()) != NULL)
		{
			last_nodes = nodes + q_nodes;
			make_move(move->move);

			/* Set up the iterative search stack and call the search. */
			if (alpha > -MATE)
			{
				/* Zero window on moves after the first. */
				search_call(&board.search_stack[0], FALSE, current_depth - PLY, 1, -alpha - 1, -alpha,
					&board.move_stack[0], NODE_PV, SEARCH_RETURN);
				value = -search(&board.search_stack[1]);
				/* research */
				if (value > alpha)
				{
					search_call(&board.search_stack[0], FALSE, current_depth - PLY, 1, -beta, -alpha,
						&board.move_stack[0], NODE_PV, SEARCH_RETURN);
					value = -search(&board.search_stack[1]);
				}
			}	
			else
			{
				search_call(&board.search_stack[0], FALSE, current_depth - PLY, 1, -beta, -alpha,
					&board.move_stack[0], NODE_PV, SEARCH_RETURN);
				value = -search(&board.search_stack[1]);
			}	
			unmake_move();
			if (stop_search())
				break;

			sum_stats();
			move->nodes = nodes + q_nodes - last_nodes;
			move->last_score = value;
	//		move->pv[0] = move->move;
	//		move->pv[1] = 0;
			//	copy_pv(move->pv, board.pv_stack[1]);
	//			display_search_line(FALSE, move->pv, current_depth, value);

			/* Update and display the new PV. */
			if (value > alpha)
			{
				alpha = value;
				move->score = ++root_pv_counter;
				best_score_by_depth[current_depth / PLY] = alpha;
				pv_changes_by_depth[current_depth / PLY]++;
				board.pv_stack[0][0] = move->move;
				copy_pv(board.pv_stack[0] + 1, board.pv_stack[1]);
				copy_pv(move->pv, board.pv_stack[0]);
				display_search_line(FALSE, board.pv_stack[0], current_depth, alpha);
			}
		}
		sum_stats();
		display_search_line(TRUE, board.pv_stack[0], current_depth, alpha);
		if (stop_search())
			break;
		/* If we're in EPD testing mode, check the move found against the solution(s). */
		if (engine_state == TESTING && check_solutions(board.pv_stack[0][0], current_depth))
			break;
	}
	if (engine_state == INFINITE)
		while (!stop_search())
			;

	/* Determine the best score, for the next search. */
	if (best_score_by_depth[current_depth / PLY] == -MATE)
		last_root_score = best_score_by_depth[(current_depth / PLY) - 1];
	else
		last_root_score = best_score_by_depth[current_depth / PLY];
	
#ifdef SMP
	stop_child_processors();
#endif

	sum_stats();
	if (protocol != XBOARD && protocol != UCI && post)
	{
		print("search: nodes=%L q=%.1f%% time=%T nps=%i\n", nodes + q_nodes, (float)100.0 * q_nodes / (nodes + q_nodes),
			time_used(), (int)(1000 * (nodes + q_nodes) / time_used()));
		print("        fail highs=%-10L pv nodes=%-10L fail lows=%-10L\n", fail_high_nodes, pv_nodes, fail_low_nodes);
		print("        fh first=%3.1f%%        pv first=%.1f%%\n", (float)100.0 * fail_high_first / fail_high_nodes,
			(float)100.0 * pv_first / pv_nodes);
		print("ext:    check=%i one-rep=%i threat=%i passed-pawn=%i\n", check_extensions_done,
			one_rep_extensions_done, threat_extensions_done, passed_pawn_extensions_done);
		print("hash:   hits=%3.1f%% cutoffs=%3.1f%% full=%3.1f%% pawn=%3.1f%% eval=%3.1f%% qsearch=%3.1f%%\n",
			(float)100.0 * hash_hits / hash_probes, (float)100.0 * hash_cutoffs / hash_probes,
			(float)100.0 * hash_entries_full / (hash_size * HASH_SLOT_COUNT), (float)100.0 *
			pawn_hash_hits / pawn_hash_probes, (float)100.0 * eval_hash_hits / eval_hash_probes,
			(float)100.0 * qsearch_hash_hits / qsearch_hash_probes);
#ifdef SMP
		print("smp:    splits=%i stops=%i/%.1f%% nodes: (", smp_data->splits_done, smp_data->stops_done, (float)100.0 *
			smp_data->stops_done / smp_data->splits_done);
		for (x = 0; x < process_count; x++)
		{
			print("%L", smp_block[x].nodes + smp_block[x].q_nodes);
			if (x < process_count - 1)
				print(", ");
		}
		print(") idle=%T (%.2f%%)\n", idle_time, (float)100.0 * idle_time / ((float)process_count * time_used()));
#endif
	}
	else if (ics_mode)
	{
		print("tellall depth=%i/%i nodes=%L(q=%.1f%%) time=%T nps=%i score=%V pv=%lM\n", current_depth / PLY, max_depth_reached,
			nodes + q_nodes, (float)100.0 * q_nodes / (nodes + q_nodes), time_used(), last_root_score,
			(int)(1000 * (nodes + q_nodes) / time_used()), board.pv_stack[0]);
#ifdef SMP
//		print(" splits=%i stops=%i/%.1f%%", smp_data->splits_done, smp_data->stops_done, (float)100.0 *
//			smp_data->stops_done / smp_data->splits_done);
#endif
		print("\n");
	}
	/* If this was a normal search, i.e. not a ponder search, set the ponder move. */
	if (engine_state == NORMAL)
		ponder_move = MOVE_COMPARE(board.pv_stack[0][1]);
	search_age++;
}

/**
sum_stats():
If we are using multiple processors, then add the stats from each processor into the global stats.
Created 083006; last modified 092906
**/
void sum_stats(void)
{
#ifdef SMP
	int x;

	nodes = q_nodes = 0;
	idle_time = 0;
	for (x = 0; x < process_count; x++)
	{
		nodes += smp_block[x].nodes;
		q_nodes += smp_block[x].q_nodes;
		idle_time += smp_block[x].idle_time;
		if (smp_block[x].last_idle_time)
			idle_time += get_time() - smp_block[x].last_idle_time;
	}
#endif
}

/**
init_stats():
Initialize all of the statistics used during the search.
Created 092906; last modified 011908
**/
void init_stats(void)
{
	int x;

#ifdef SMP
	smp_data->stop = FALSE;
	smp_data->splits_done = 0;
	smp_data->stops_done = 0;
	for (x = 0; x < process_count; x++)
	{
		smp_block[x].nodes = 0;
		smp_block[x].q_nodes = 0;
		smp_block[x].idle_time = 0;
		if (x != 0)
			smp_block[x].last_idle_time = get_time();
	}
#endif
	nodes = 0;
	q_nodes = 0;
	max_depth_reached = 0;
	nodes_until_check = nodes_per_check;
	pv_nodes = 0;
	fail_high_nodes = 0;
	fail_low_nodes = 0;
	pv_first = 0;
	fail_high_first = 0;
	hash_probes = 0;
	hash_hits = 0;
	hash_cutoffs = 0;
	pawn_hash_probes = 0;
	pawn_hash_hits = 0;
	eval_hash_probes = 0;
	eval_hash_hits = 0;
	qsearch_hash_probes = 0;
	qsearch_hash_hits = 0;
	check_extensions_done = 0;
	one_rep_extensions_done = 0;
	threat_extensions_done = 0;
	singular_extensions_done = 0;
	
	for (x = 0; x < MAX_PLY; x++)
	{
		best_score_by_depth[x] = -MATE;
		pv_changes_by_depth[x] = 0;
	}
}
